#!/usr/bin/env python3

import copy
import json
import os
import pprint
import time
import traceback
from dataclasses import asdict, dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Sequence, Set, Union

from cmk.utils.type_defs import HostName, UserId

from cmk.gui import sites
from cmk.gui.breadcrumb import make_current_page_breadcrumb_item, make_topic_breadcrumb
from cmk.gui.hooks import request_memoize
from cmk.gui.htmllib.header import make_header
from cmk.gui.htmllib.html import html
from cmk.gui.http import request
from cmk.gui.i18n import _, _l
from cmk.gui.log import logger
from cmk.gui.main_menu import mega_menu_registry
from cmk.gui.node_visualization import (
    _get_topology_configuration,
    ABCTopologyPage,
    Meshes,
    Topology,
    topology_registry,
)
from cmk.gui.nodevis_lib import TopologyConfiguration
from cmk.gui.page_menu import PageMenu
from cmk.gui.pages import page_registry
from cmk.gui.pagetypes import PagetypeTopics
from cmk.gui.type_defs import ColumnName, Row, VisualLinkSpec, ColumnSpec
from cmk.gui.utils.theme import theme
from cmk.gui.utils.urls import makeuri_contextless
from cmk.gui.view_utils import CellSpec
from cmk.gui.views import Icon, icon_and_action_registry, painter_registry, multisite_builtin_views, \
    builtin_views
from cmk.gui.views.painter.v0.base import Cell, Painter


@dataclass
class CMCHostObject:
    hostname: HostName
    alias: str
    icon_image: Optional[str]
    address: Optional[str]
    state: int
    has_been_checked: int


@dataclass
class CMCInterfaceObject:
    host_state: int
    service_state: int
    num_services_warn: int
    num_services_crit: int
    icon_image: str


demo_mode = False
obfuscate_hostnames = False


class CMCHostsProvider:
    def __init__(self) -> None:
        self._cmc_hosts = {}

    def fetch_host_info(self, hostnames: Set[HostName]):
        unfetched_hostnames = hostnames - set(self._cmc_hosts)
        if not unfetched_hostnames:
            return

        # The core is unable to handle big filters
        # Simply fetch everything if too much hosts are queried
        if len(hostnames) > 1000:
            # TODO: remove
            #            hostname_filter = f"Filter: name ~ ^({'|'.join(hostnames)})$"
            hostname_filter = ""
        else:
            hostname_filter = "\n".join("Filter: name = %s" % x for x in hostnames)
            hostname_filter += "\nOr: %d" % len(hostnames)
        query = (
            "GET hosts\nColumns: name alias icon_image address state has_been_checked\n%s"
            % hostname_filter
        )

        for hostname, alias, icon_image, address, state, has_been_checked in sites.live().query(
            query
        ):
            self._cmc_hosts[hostname] = CMCHostObject(
                hostname, alias, icon_image, address, state, has_been_checked
            )

    @property
    def cmc_hosts(self) -> Dict[HostName, CMCHostObject]:
        return self._cmc_hosts


@request_memoize()
def _cmc_hosts_provider():
    return CMCHostsProvider()


@page_registry.register_page("network_topology")
class NetworkTopologyPage(ABCTopologyPage):
    def show_topology(self) -> None:
        visual_spec = self.visual_spec()
        breadcrumb = make_topic_breadcrumb(
            mega_menu_registry.menu_monitoring(),
            PagetypeTopics.get_topic(visual_spec["topic"]).title(),
        )
        breadcrumb.append(make_current_page_breadcrumb_item(str(visual_spec["title"])))
        page_menu = PageMenu(breadcrumb=breadcrumb)
        self._extend_display_dropdown(page_menu, visual_spec["name"])
        make_header(html, str(visual_spec["title"]), breadcrumb, page_menu)
        # TODO: remove
        #        nodevis_javascript = (
        #            Path(cmk.utils.paths.local_web_dir) / "htdocs" / "js" / "network_visualization_min.js"
        #        ).read_text()
        #        nodevis_javascript = (
        #            Path(cmk.utils.paths.web_dir) / "htdocs" / "js" / "network_visualization_min.js"
        #        ).read_text()
        html.write_html(
            f'<script type="text/javascript" src={"js/network_visualization_min.js"}></script>'
        )
        self.show_topology_content()

    def show_topology_content(self) -> None:
        div_id = "node_visualization"
        html.div("", id_=div_id)

        search_frontend_settings = not request.has_var("topology_frontend_configuration")
        topology_configuration, _query_hash = _get_topology_configuration(
            self.visual_spec()["name"], self._get_overlays_config()
        )

        html.javascript(
            "topology_instance = new cmk.nodevis_network.TopologyVisualization(%s,%s);"
            % (json.dumps(div_id), json.dumps(topology_configuration.type))
        )

        html.javascript(
            "topology_instance.show_topology(%s,%s)"
            % (topology_configuration.frontend.to_json(), json.dumps(search_frontend_settings))
        )

    @classmethod
    def visual_spec(cls):
        return {
            "topic": "overview",
            "title": _("Network topology"),
            "name": "layer_2_topology",
            "sort_index": 50,
            "is_show_more": False,
            "icon": "layer_2_topology",
            "hidden": False,
            "single_infos": [],
            "context": {},
            "link_from": {},
            "add_context_to_title": True,
        }

    def _get_overlays_config(self):
        return {"network_overlay": {"active": True, "configurable": False}}


class NetworkTopology(Icon):
    @classmethod
    def ident(cls):
        return "layer_2_topology"

    @classmethod
    def title(cls):
        return _("Layer2 Topology")

    def host_columns(self):
        return ["name"]

    def default_sort_index(self):
        return 51

    def render(self, what, row, tags, custom_vars):
        url = makeuri_contextless(
            request, [("host_regex", f"{row['host_name']}$")], filename="network_topology.py"
        )
        return "aggr", _("Network layer topology"), url


icon_and_action_registry.register(NetworkTopology)


class InterfaceDisplayMode(Enum):
    off = "off"
    only_problems = "only_problems"
    all = "all"


class NetworkLinkType(Enum):
    host_host = "host_host"
    host_if = "host_if"
    if_if = "if_if"


@dataclass(kw_only=True)
class NetworkTopologyNode:
    name: HostName
    hostname: HostName
    address: str
    alias: str
    state: int
    icon_image: str
    incoming: set
    outgoing: set


@dataclass(kw_only=True)
class NetworkTopologyHost(NetworkTopologyNode):
    type: str = "host"


@dataclass(kw_only=True)
class NetworkTopologyInterface(NetworkTopologyNode):
    type: str = "interface"


def create_topology_dict(network_topology_host: NetworkTopologyHost):
    result = {}
    for field in network_topology_host.__dataclass_fields__:
        value = getattr(network_topology_host, field)
        if field in ("incoming", "outgoing"):
            result[field] = list(value)
        else:
            result[field] = value
    return result


def _network_data_path():
    return Path("~/var/topology_data").expanduser()


@request_memoize()
def get_network_data(datasource_id: str):
    return json.loads(open(_network_data_path() / datasource_id / "network_data.json").read())


@dataclass
class LLDPLinkInfo:
    source: str
    target: str
    config: dict[str, Any]

    def to_json(self):
        return {
            "source": self.source,
            "target": self.target,
            "config": self.config,
        }

    def __hash__(self):
        tokens = tuple(sorted([self.source, self.target]))
        return ("%s_%s" % tokens).__hash__()

    def id(self) -> tuple[str]:
        return tuple(sorted([self.source, self.target]))


def map_host_state_to_service_state(info: dict[str, Any], host_info: dict[str, Any]) -> int:
    if info["node_type"] in ["topology_center", "topology_site"]:
        return 0
    if not host_info["has_been_checked"]:
        if demo_mode:
            return 0
        return -1
    if host_info["state"] == 1:
        return 2
    return 0


def _get_default_interface_display_mode(mesh_root_settings) -> InterfaceDisplayMode:
    if (network_info := mesh_root_settings.get("network_info")) is None:
        return InterfaceDisplayMode.only_problems

    if (if_display_mode := network_info.get("if_display_mode")) is None:
        return InterfaceDisplayMode.only_problems

    return InterfaceDisplayMode(if_display_mode)


class LLDPTopology(Topology):
    def __init__(
        self, topology_config: TopologyConfiguration, use_datasource: Optional[str] = None
    ):
        super().__init__(topology_config)
        self._changed_names = {}

        self._network_info: dict[HostName, dict[str, Any]] = {}
        self._explicit_interface_settings: dict[str, InterfaceDisplayMode] = {}
        self._styles = []

        ds_config = topology_config.frontend.datasource_configuration
        self._datasource_id = use_datasource if use_datasource else ds_config.reference

        self._compare_to_topology: Optional[LLDPTopology] = None
        if not use_datasource and ds_config.reference != ds_config.compare_to:
            # Create a topology with the same settings but a different datasource
            # This topology is used later on to compute the differences
            self._compare_to_topology = LLDPTopology(
                topology_config, use_datasource=ds_config.compare_to
            )

        self._default_if_display_mode = _get_default_interface_display_mode(
            self._settings.frontend.custom_node_settings.get("mesh_root", {})
        )

        for node_id, settings in self._settings.frontend.custom_node_settings.items():
            if value := settings.get("network_info"):
                if if_setting := value.get("if_display_mode"):
                    self._explicit_interface_settings[node_id] = InterfaceDisplayMode(if_setting)

        self._hosts_in_mesh = 0

    @classmethod
    def ident(cls):
        return "layer_2_topology"

    def title(self):
        return _("Network topology")

    def _fetch_data_for_hosts(self, hostnames: Set[HostName]) -> Sequence[NetworkTopologyHost]:
        _cmc_hosts_provider().fetch_host_info(hostnames)

        network_data = get_network_data(self._datasource_id)
        for hostname in hostnames:
            cmc_host = _cmc_hosts_provider().cmc_hosts.get(hostname)
            if not cmc_host:
                logger.warning(f"hostname not known by cmc {hostname}")
                continue

            network_host = NetworkTopologyHost(
                name=hostname,
                hostname=hostname,
                address=cmc_host.address,
                alias=cmc_host.alias,
                state=map_host_state_to_service_state(
                    {"node_type": ""}, asdict(cmc_host)
                ),  # TODO: cleanup node type
                icon_image=cmc_host.icon_image,
                incoming=set(),
                outgoing=set(),
            )

            if network_info := network_data.get(hostname):
                self._network_info[hostname] = network_info
                for if_name, target_info in network_info.get("connections").items():
                    target_host, _target_if, _datasource_type = target_info
                    network_host.outgoing.add(target_host)

            yield create_topology_dict(network_host)

    def _grow(self):
        self._fetch_root_nodes()
        self._growth_to_depth()

    def _obfuscate_hostname(self, node_name):
        if name := self._changed_names.get(node_name):
            return name
        self._changed_names[node_name] = f"Host_{len(self._changed_names)}"
        return self._changed_names[node_name]

    def get_info_for_node(self, hostname, mesh):
        info = super().get_info_for_node(hostname, mesh)
        host_info = self._known_nodes[hostname]
        info.update(host_info)
        info = copy.deepcopy(info)
        del info["incoming"]
        del info["outgoing"]
        info["date"] = time.strftime("%H:%M:%S", time.localtime())

        if info["type"] == "interface":
            info["node_type"] = "network_interface"
            info["name"] = hostname.split("_")[1]
            info["show_text"] = True
            info["id"] = hostname
        else:
            info["node_type"] = "network_host"
            info["network_data"] = get_network_data(self._datasource_id).get(hostname)
            info["id"] = hostname
            if obfuscate_hostnames:
                info["name"] = self._obfuscate_hostname(hostname)

        if info.get("icon_image"):
            info["icon_image"] = theme.detect_icon_path(info["icon_image"], "icon_")

        if style := self._get_style(info):
            self._styles.append(style)
        return info

    def _get_style(self, info):
        #        if available_config := self._settings.frontend.node_style_config.get(info["hostname"]):
        #            if not (available_config["type"] == "fixed" and info["hostname"] in self._hosts_with_interfaces):
        #                return available_config
        #
        #        # If the mesh root has an explicit config, there is no need to add default configs to its children
        #        if self._settings.frontend.node_style_config.get("mesh_root"):
        #            return

        hostname = info["hostname"]

        def _default_force_style():
            return {"type": "none", "style_config": {}}

        def _default_hierarchy_style():
            return {
                "type": "hierarchy",
                "options": {
                    "box_leaf_nodes": False,
                    "layer_height": 150,
                    "node_size": 30,
                    "rotation": 0,
                },
                "matcher": {"id": {"value": info["id"]}},
            }

        def _default_radial_style():
            return {
                "type": "radial",
                "options": {
                    "radius": 10,
                    "degree": 360,
                    "rotation": 90,
                },
                "matcher": {"id": {"value": info["id"]}},
            }

        explicit_settings = self._explicit_interface_settings.get(hostname)
        if not explicit_settings:
            return None

        if explicit_settings.value != InterfaceDisplayMode.all:
            return

        preferred_style = {
            "F": _default_force_style(),
            "H": _default_hierarchy_style(),
            "R": _default_radial_style(),
        }[
            self._settings.frontend.custom_node_settings.get(hostname, {}).get(
                "preferred_style", "H"
            )
        ]
        # current_style = self._settings.frontend.node_style_config.get(hostname)
        current_style = None
        if (
            current_style
            and preferred_style
            and current_style.get("type") == preferred_style["type"]
        ):
            return {
                "type": current_style["type"],
                "style_config": current_style["options"],
            }

        if preferred_style is None:
            return _default_hierarchy_style()

        return preferred_style

    def compute(self):
        # Combine all nodes into one big mesh, even if they are not connected
        meshes = super().compute()
        merged_meshes = set()
        for mesh in meshes:
            merged_meshes.update(mesh)
        return [merged_meshes]

    def get_mesh_root(self, mesh):
        return "mesh_root"

    def generate_mesh_configuration(self, mesh):
        try:
            return self._generate_mesh_configuration(mesh)
        except Exception as e:
            logger.warning(traceback.format_exception(e))

    def _generate_mesh_configuration(self, mesh):
        # Pick root host
        actual_hosts = {}
        for hostname in mesh:
            tokens = hostname.split("_")
            host = actual_hosts.setdefault(tokens[0], [])
            if len(tokens) == 2:
                host.append(hostname)

        mesh_root = self.get_mesh_root(mesh)
        mesh_root_info = {
            "hostname": mesh_root,
            "name": mesh_root,
            "id": mesh_root,
            "state": 0,
            "node_type": "mesh_root",
            "children": [],
            "custom_node_settings": self._settings.frontend.custom_node_settings.get(
                "mesh_root", {}
            ),
        }

        mesh_links: set[LLDPLinkInfo] = set([])

        for hostname in sorted(mesh):
            host_info = self.get_info_for_node(hostname, mesh)

            if hostname in actual_hosts:
                mesh_root_info["children"].append(host_info)
                network_info = self._network_info.get(hostname)
                if network_info:
                    if interfaces := sorted(network_info["connections"].keys()):
                        host_info["children"] = []
                        for if_name in interfaces:
                            if_name = if_name.strip()
                            if_node_name = f"{hostname}_{if_name}"
                            if if_node_name in mesh:
                                host_info["children"].append(
                                    self.get_info_for_node(if_node_name, mesh)
                                )
                        host_info["children"].sort(key=lambda info: info["hostname"])
                if len(host_info.setdefault("children", [])) == 0:
                    # TODO: cleanup
                    # Add fake child, 'fixes' layouting problems in radial layout
                    host_info["children"].append(
                        {
                            "hostname": "hidden",
                            "name": f"hidden_{hostname}",
                            "id": f"hidden_{hostname}",
                            "state": 0,
                            "node_type": "hidden_network",
                            "invisible": True,
                            "children": [],
                        }
                    )

            for child in self.get_host_incoming(hostname):
                if child in mesh:
                    mesh_links.add(self._compute_link(host_info, child, hostname))

            for child in self.get_host_outgoing(hostname):
                if child in mesh:
                    mesh_links.add(self._compute_link(host_info, child, hostname))

        if self._compare_to_topology:
            mesh_configuration = self._compare_to_topology.generate_mesh_configuration(
                self._compare_to_topology.compute()[0]
            )
            self._integrate_differences(mesh_configuration, mesh, mesh_links, mesh_root_info)

        response = {
            "layout": {
                "config": {
                    "line_config": {
                        "style": "round",
                        "dashed": False,
                    },
                    "style_configs": self._settings.frontend.style_configs,
                }
            },
            "type": "topology",
            "hierarchy": mesh_root_info,
            "links": list(mesh_links),
        }

        if not self._settings.frontend.style_configs:
            response["layout"]["config"]["style_configs"] = self._generate_initial_style_config(
                self.get_mesh_root(mesh), len(mesh)
            )

        # TODO: this doesn't belong here..
        #       requires a refactoring oif the node_visualiziation.py
        self._settings.frontend.datasource_configuration.available_datasources = sorted(
            os.listdir(_network_data_path())
        )

        #        logger.warning("RESPONSE READY")
        #        logger.warning(f"response {pprint.pformat(response)}")
        return response

    def _generate_initial_style_config(self, hostname, mesh_size):
        root_style_config = [
            {
                "matcher": {"hostname": {"value": self.get_mesh_root({})}},
                "options": {
                    #                    "degree": 360 if mesh_size > 2 else 120,
                    "degree": 360,
                    "detach_from_parent": False,
                    "include_parent_rotation": False,
                    "radius": 100 + self._hosts_in_mesh * 5,
                    "rotation": 0,
                },
                "position": {"x": 50, "y": 50},
                "type": "radial",
            }
        ]

        root_style_config.extend(self._styles)
        return root_style_config

    def _integrate_differences(
        self, other_mesh_configuration: dict[str, Any], mesh, mesh_links, mesh_root_info
    ):
        other_links = other_mesh_configuration["links"]
        other_root_info = other_mesh_configuration["hierarchy"]

        # Find new and vanished links for
        reference_links = {x.id(): x for x in mesh_links}
        other_links = {x.id(): x for x in other_links}
        other_node_lookup = {}
        for x in other_root_info["children"]:
            other_node_lookup[x["id"]] = x
            for child in x.get("children", []):
                other_node_lookup[child["id"]] = child

        for only_in_reference in set(reference_links) - set(other_links):
            reference_links[only_in_reference].config["link_info"] = "Only in reference"
            reference_links[only_in_reference].config["css"] = "only_in_reference"
        for only_in_other in set(other_links) - set(reference_links):
            other_links[only_in_other].config["link_info"] = "Missing in reference"
            other_links[only_in_other].config["css"] = "missing_in_reference"
            mesh_links.add(other_links[only_in_other])
            # TODO: Highly experimental!
            logger.warning(f"only in other {only_in_other}")

            for other_node_id in only_in_other:
                if other_node_id not in mesh:
                    if "_" in other_node_id:
                        # An interface node. Create an additional host node
                        # if this host is unknown in the reference mesh
                        hostname, ifname = other_node_id.split("_", 1)
                        for child in mesh_root_info["children"]:
                            if child["hostname"] == hostname:
                                logger.warning(f"use existing host {hostname}")
                                host_ref = child
                                break
                        else:
                            host_ref = other_node_lookup[hostname]

                            logger.warning(f"add missing host {hostname}")
                            mesh_root_info["children"].append(host_ref)
                            host_ref["css"] = "missing"

                        other_node_lookup[other_node_id]["css"] = "missing"
                        other_if_node = other_node_lookup[other_node_id]
                        other_if_node["css"] = "missing"
                        for if_child in host_ref["children"]:
                            logger.warning(f"checking {if_child['hostname']} with {ifname}")
                            if if_child["hostname"] == other_node_id:
                                break
                        else:
                            logger.warning(f"add missing interface {other_node_id}")
                            host_ref["children"].append(other_if_node)
                    else:
                        other_node_lookup[other_node_id]["css"] = "missing"
                        for child in mesh_root_info["children"]:
                            if child["hostname"] == other_node_id:
                                break
                        else:
                            logger.warning(f"add missing simple host {other_node_id}")
                            mesh_root_info["children"].append(other_node_lookup[other_node_id])
                    mesh.add(other_node_id)

    def _compute_link(
        self, host_info: dict[str, Any], source: str, target: str, is_crit: bool = False
    ) -> LLDPLinkInfo:
        """source target directory doesn't matter. sort them to avoid duplicates"""
        nodes = sorted([source, target])
        # nodes = [source, target]
        link_type = self._get_link_type(nodes[0], nodes[1])

        css_state = 0
        if link_type != NetworkLinkType.host_host:
            for node in nodes:
                if if_info := self._core_interface_info.get(tuple(node.split("_"))):
                    css_state = max(css_state, if_info.service_state)

        css = (
            {
                0: "",
                1: "warning",
                2: "critical",
                3: "unknown",
            }[css_state],
        )

        return LLDPLinkInfo(
            source=nodes[0],
            target=nodes[1],
            config={
                "type": link_type.value,
                "css": css,
            },
        )

    def _get_link_type(self, source, target) -> NetworkLinkType:
        if "_" not in source and "_" not in target:
            return NetworkLinkType.host_host
        if "_" in source and "_" in target:
            return NetworkLinkType.if_if
        return NetworkLinkType.host_if

    def _if_description_split(self, name: str) -> tuple[Union[str, int]]:
        tokens = []
        for token in name.split("/"):
            try:
                tokens.append(int(token))
            except ValueError:
                tokens.append(token)

        return tuple(tokens)

    def _get_core_interface_info(self, hostnames) -> dict[(str, str), CMCInterfaceObject]:
        """Fetch interface states and update nodes"""
        query = [
            "GET services",
            "Columns: host_name host_state service_description state has_been_checked host_num_services_warn host_num_services_crit icon_image",
        ]
        # for hostname in hostnames:
        #    query.append(
        #        f"Filter: host_name = {hostname}\nFilter: description ~ Interface \nAnd: 2"
        #    )
        # query.append(f"Or: {len(hostnames)}")
        query.append(f"Filter: host_name ~ ^({'|'.join(hostnames)})$")
        query.append(f"Filter: description ~ Interface")

        interface_states = {}
        for (
            host_name,
            host_state,
            service_description,
            service_state,
            has_been_checked,
            num_services_warn,
            num_services_crit,
            icon_image,
        ) in sites.live().query("\n".join(query)):
            if not has_been_checked:
                service_state = -1
            interface_states[(host_name, service_description[10:])] = CMCInterfaceObject(
                host_state,
                service_state,
                num_services_warn,
                num_services_crit,
                icon_image,
            )

        return interface_states

    def _create_connection_map(self, meshes) -> Dict[str, set[str]]:
        connection_map = {}
        for node_name in meshes:
            connection_map.setdefault(node_name, set())
            for target_host in self._known_nodes[node_name].get("outgoing"):
                connection_map[node_name].add(target_host)
                connection_map.setdefault(target_host, set()).add(node_name)
        return connection_map

    def _postprocess_meshes(self, meshes: Meshes) -> Meshes:
        """Fetch interface states and update nodes"""
        hosts_to_fetch = []
        merged_mesh = set()
        for sub_mesh in meshes:
            merged_mesh.update(sub_mesh)

        if not merged_mesh:
            return [merged_mesh]

        hosts_to_fetch.extend(merged_mesh)

        self._hosts_in_mesh = len(hosts_to_fetch)

        connection_map = self._create_connection_map(merged_mesh)
        self._core_interface_info = self._get_core_interface_info(set(hosts_to_fetch))

        # Cleanup outgoing nodes
        for node_name in list(merged_mesh):
            # Remove host-host links to remote hosts if an interface already connects to it
            # Might be a bit misleading, if multiple interfaces connect to the same remote host
            # and only one interface with an error is shown
            # TODO: only cut the host-host link if all interfaces to the remote host are visible
            for if_node in self._create_interfaces(node_name):
                self._known_nodes.update({if_node.hostname: create_topology_dict(if_node)})
                if_target = list(if_node.outgoing)[0]
                target_host = if_target.split("_", 1)[0]
                if target_host in self._known_nodes[node_name]["outgoing"]:
                    self._known_nodes[node_name]["outgoing"].remove(target_host)
                merged_mesh.add(if_node.hostname)

                # Remove links from other hosts
                for target_host in connection_map[node_name]:
                    if (
                        target_host in self._known_nodes
                        and node_name in self._known_nodes[target_host]["outgoing"]
                    ):
                        self._known_nodes[target_host]["outgoing"].remove(node_name)

            self._known_nodes[node_name]["has_no_parents"] = (
                self._known_nodes[node_name]["type"] == "host"
            )

        return [merged_mesh]

    def _create_interfaces(self, node_name) -> Iterable[NetworkTopologyHost]:
        if node_name not in self._network_info:
            return []

        for if_name, target_info in self._network_info[node_name]["connections"].items():
            if interface_info := self._core_interface_info.get((node_name, if_name)):
                self._known_nodes[node_name]["num_services_warn"] = interface_info.num_services_warn
                self._known_nodes[node_name]["num_services_crit"] = interface_info.num_services_crit
                icon_image = interface_info.icon_image
                if_state = interface_info.service_state
            else:
                self._known_nodes[node_name]["num_services_warn"] = 0
                self._known_nodes[node_name]["num_services_crit"] = 0
                icon_image = ""
                if_state = -1

            if_name = if_name.strip()

            if demo_mode and if_state == -1:
                if_state = 0

            display_mode = (
                self._explicit_interface_settings.get(node_name) or self._default_if_display_mode
            )

            if display_mode == InterfaceDisplayMode.off:
                continue

            if if_state in (0, -1) and display_mode == InterfaceDisplayMode.only_problems:
                continue

            if_node = NetworkTopologyInterface(
                name=if_name,
                hostname=f"{node_name}_{if_name}",
                address="127.0.0.1",
                alias=if_name,
                state=if_state,
                icon_image=icon_image,
                incoming=set([node_name]),
                outgoing=set(),
            )
            target_host, target_if_name, _datasource_type = target_info
            display_mode = (
                self._explicit_interface_settings.get(target_host) or self._default_if_display_mode
            )

            if target_interface_info := self._core_interface_info.get(
                (target_host, target_if_name)
            ):
                target_if_state = target_interface_info.service_state
            else:
                target_if_state = -1

            if demo_mode and target_if_state == -1:
                target_if_state = 0

            if display_mode == InterfaceDisplayMode.all or (
                target_if_state > 0 and display_mode == InterfaceDisplayMode.only_problems
            ):
                if_node.outgoing.add(f"{target_host}_{target_if_name}")
            else:
                if_node.outgoing.add(target_host)
            yield if_node


topology_registry.register(LLDPTopology)


class PainterLLDPInfo(Painter):
    @property
    def ident(self) -> str:
        return "network_info"

    def title(self, cell: Cell) -> str:
        return _("Network connection")

    def short_title(self, cell: Cell) -> str:
        return _("Network connection")

    @property
    def columns(self) -> Sequence[ColumnName]:
        return ["service_description", "host_name"]

    def render(self, row: Row, cell: Cell) -> CellSpec:
        network_data = get_network_data(request.get_str_input_mandatory("datasource", "default"))
        hostname = row["host_name"]
        service_description = row["service_description"]
        if not service_description.startswith("Interface "):
            return "", ""

        service_description = service_description[10:]

        if (host_info := network_data.get(hostname)) is None:
            return "", "No network connection data available"

        if service_info := host_info["connections"].get(service_description):
            remote_hostname = service_info[0]
            if obfuscate_hostnames:
                remote_hostname = "-censored-"
            return "", _("Connected to %s / %s") % (remote_hostname, service_info[1])

        if service_description in host_info["interfaces"]:
            return "", _("Interface known, but not connected")

        return "", _("Interface not found in network connection data")


painter_registry.register(PainterLLDPInfo)

_service_view_painters = [
    ColumnSpec(name="service_state"),
    ColumnSpec(
        name="service_description",
        link_spec=VisualLinkSpec(type_name="views", name="service"),
    ),
    ColumnSpec(name="service_icons"),
    ColumnSpec(name="svc_plugin_output"),
    ColumnSpec(name="svc_state_age"),
    ColumnSpec(name="svc_check_age"),
    ColumnSpec(name="perfometer"),
    ColumnSpec(name="network_info"),
]

multisite_builtin_views.update(
    {
        "network_hover_service": {
            "add_context_to_title": True,
            "browser_reload": 30,
            "column_headers": "pergroup",
            "context": {},
            "datasource": "services",
            "description": "Information shown when hovering over a topology service noden",
            "force_checkboxes": False,
            "group_painters": [],
            "hidden": True,
            "hidebutton": False,
            "icon": None,
            "owner": UserId.builtin(),
            "is_show_more": False,
            "layout": "dataset",
            "link_from": {},
            "mobile": False,
            "mustsearch": False,
            "name": "network_hover_service",
            "num_columns": 1,
            "packaged": False,

            "painters": _service_view_painters,
            "play_sounds": False,
            "public": True,
            "single_infos": ["service", "host"],
            "sort_index": 99,
            "sorters": [],
            "title": "Service",
            "topic": "other",
            "user_sortable": True,
        }
    }
)
